<?php

class NitroCoreRelation_OneToMany extends NitroRelation
{
	protected $_old_related = array();
	protected $_added = array();
	protected $_removed = array();
	
	/**
	 * Gets the related entity/ies
	 * 
	 * @param $meth_name	The relationship method name to get
	 * @param $params		Optionally sets where, order_by, group_by, limit_value and limit_offset to filter collection
	 * @return NitroCollection
	 */
    public function getRelated( $meth_name, $params = array() )
	{
		$custom = ! empty( $params );
		
		if ( ! @$this->_related[ $meth_name ] || $custom )
		{
			$entity = $this->_map[ $meth_name ]['entity'];
			$fields = (array) $this->_map[ $meth_name ]['fields'];
			$dbname = $entity::DB()->database;
			$table = $entity::table();
			
			if ( ( $pks = $this->_main->PK() ) )
			{
				// Build where for PKs
				$pks = (array) $pks;
				foreach( $fields as $idx => $field )
					$pk_where[ preg_match("/\./i",$field) ? $field : "$dbname.$table.$field" ] = $pks[ $idx ];
				
				$q = $entity::DB()
					->select( $entity::PKFields() )
					->where( $pk_where );
				
				$entity::parentJoin(); // Join with its parent table
				
				// Apply join from mapping
				if ( $this->_map[ $meth_name ]['params']['join'] )
					foreach ( (array)$this->_map[ $meth_name ]['params']['join'] as $key => $val )
						if ( is_string($val) )
							$q->join( $key, $val );
						else
							$q->join( $key, $val[1], $val[2] );
				// Apply join from method call
				if ( $params["join"] )
					foreach ( (array)$params["join"] as $key => $val )
						if ( is_string($val) )
							$q->join( $key, $val );
						else
							$q->join( $key, $val[1], $val[2] );
				
				// Apply where from mapping
				if ( $this->_map[ $meth_name ]['params']['where'] )
					foreach ( (array)$this->_map[ $meth_name ]['params']['where'] as $key => $val )
						if ( is_string($key) )
							$q->where( array( $key => $val ) );
						else
							$q->where( $val, NULL, FALSE );
				// Apply where from method call
				if ( $params["where"] )
					foreach ( (array)$params["where"] as $key => $val )
						if ( is_string($key) )
							$q->where( array( $key => $val ) );
						else
							$q->where( $val, NULL, FALSE );
				
				// Apply order by from mapping
				if ( @$this->_map[ $meth_name ]['params']['order_by'] )
					$q->order_by( $this->_map[ $meth_name ]['params']['order_by'] );
				// Apply order by from method call
				if ( $params["order_by"] )
					$q->order_by( $params["order_by"] );
				
				// Apply group by from mapping
				if ( $this->_map[ $meth_name ]['params']['group_by'] )
					$q->group_by( $this->_map[ $meth_name ]['params']['group_by'] );
				// Apply group by from method call
				if ( $params["group_by"] )
					$q->group_by( $params['group_by'] );
				
				// Apply limit from mapping
				if ( @$this->_map[ $meth_name ]['params']['limit_value'] )
					$q->limit( $this->_map[ $meth_name ]['params']['limit_value'], @$this->_map[ $meth_name ]['params']['limit_offset'] );
				// Apply limit from method call
				if ( $params["limit_value"] )
					$q->limit( $params["limit_value"], @$params["limit_offset"] );
				
				if ( ( $rs = $q->get( $table ) ) && $rs->num_rows() )
					$data = array_map( 'array_values', $rs->result_array() );
				//if ( $meth_name == "Contacts" )echo("PARENTTTTTTTTTTTT: ".$this->_main);
			}
			
			$col = new NitroCollection( $entity, (array)$data );
			if ( $custom )
				return $col;
			
			$this->_related[ $meth_name ] = $col;
			//if ( $meth_name == "Contacts" )echo("ENTRAAAAAAAAAAAAAA: $entity");
		}
		
		//if ( $meth_name == "Contacts" )var_dump($entity::DB());exit;
		
		return $this->_related[ $meth_name ];
	}
	
	/**
	 * Sets the related entity/ies
	 * 
	 * @param String $meth_name		The relationship method name to get
	 * @param NitroCollection $related
	 */
	public function setRelated( $meth_name, NitroCollection $related )
	{
		$ent = $this->_map[ $meth_name ]['entity'];
		if ( $related->className() != $ent )
			throw new NitroException("The collection must be of '$ent' entity.");
		
		// Keep track of old related collections
		$this->_old_related[ $meth_name ] = TRUE;
			
		return parent::setRelated( $meth_name, $related );
	}
	
	public function clearRelated( $meth_name )
	{
		$this->_old_related[ $meth_name ] = TRUE;
		$this->_related[ $meth_name ] = new NitroCollection( $this->_map[ $meth_name ]['entity'] );
		
		return $this->_main;
	}
	
	/**
	 * Saves related entity/ies BEFORE main entity save()
	 * 
	 * @return Bool
	 */
    public function savePre( $force_insert = FALSE )
	{
		if ( ( $pks = $this->_main->PK() ) )
		{
			$main =& $this->_main;
			
			if ( ! empty($this->_old_related) )
			{
				foreach ( $this->_old_related as $meth_name => $dummy )
				{
					$set = array();
					$where = array();
					$entity = $this->_map[ $meth_name ]['entity'];
					$fields = (array) $this->_map[ $meth_name ]['fields'];
					$dbname = $entity::DB()->database;
					$table = $entity::table();
					$pks = (array) $pks;

					foreach( $fields as $idx => $field )
					{
						$set[ "$dbname.$table.$field" ] = NULL;
						$where[ "$dbname.$table.$field" ] = $pks[ $idx ];
					}

					if ( @$this->_map[ $meth_name ]['loose'] )
						$main::DB()->set( $set )->where( $where )->update( $table );
					else if ( ( $m = $entity::find( $where ) ) )
						$m->delete();
				}
			}
			
			if ( ! empty($this->_removed) )
			{
				
				foreach ( (array)$this->_removed as $meth_name => $entities )
				{
					$set = array();
					$many_class = $this->_map[ $meth_name ]['entity'];
					$dbname = $many_class::DB()->database;
					$table = $many_class::table();
					$pkf = $many_class::PKFields();
					
					foreach( (array)$this->_map[ $meth_name ]['fields'] as $idx => $field )
						$set[ "$dbname.$table.$field" ] = NULL;
					
					foreach ( $entities as $ent_or_pk )
					{
						$is_model = $ent_or_pk instanceof NitroModel;
						$pk = (array) ( $is_model ? $ent_or_pk->PK() : $ent_or_pk );

						if ( @$this->_map[ $meth_name ]['loose'] )
						{
							$where = array();
							
							foreach( $pkf as $idx => $field )
								$where[ "$dbname.$table.$field" ] = $pk[ $idx ];
							
							$main::DB()->set( $set )->where( $where )->update( $table );
						}
						else if ( $is_model )
							$ent_or_pk->delete();
						else if ( ( $m = $many_class::findByPK( $pk ) ) )
							$m->delete();
					}
				}
			}
		}
		
		return $this->_old_related = array();
	}
	
	/**
	 * Saves related entity/ies AFTER main entity save()
	 * 
	 * @return Bool
	 */
	public function savePost( $force_insert = FALSE )
	{
		$pks = (array) $this->_main->PK();
		
		foreach ( (array)$this->_related as $meth_name => $col )
		{
			$fields = (array) $this->_map[ $meth_name ]['fields'];
			
			foreach ( (array) $this->_added[ $meth_name ] as $ent )
				foreach( $fields as $idx => $field )
					$ent->set( $field, $pks[ $idx ] );
			
			$col->save();
		}
	}
	
	public function add( $meth_name, NitroModel $ent )
	{
		$this->_added[ $meth_name ][ ] = $this->getRelated( $meth_name )->add( $ent );
		
		return $this->_main;
	}
	
	public function remove( $meth_name, $ent_or_pk )
	{
		if ( $this->getRelated( $meth_name )->remove( $ent_or_pk ) )
			$this->_removed[ $meth_name ][ ] = $ent_or_pk;
		
		return $this->_main;
	}
}